@@ -1,27 +1,52 @@
+/*
+     PLIB - A Suite of Portable Game Libraries
+     Copyright (C) 2001  Steve Baker
+ 
+     This library is free software; you can redistribute it and/or
+     modify it under the terms of the GNU Library General Public
+     License as published by the Free Software Foundation; either
+     version 2 of the License, or (at your option) any later version.
+ 
+     This library is distributed in the hope that it will be useful,
+     but WITHOUT ANY WARRANTY; without even the implied warranty of
+     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+     Library General Public License for more details.
+ 
+     You should have received a copy of the GNU Library General Public
+     License along with this library; if not, write to the Free
+     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
+ 
+     For further information visit http://plib.sourceforge.net
+
+     $Id: pu.h,v 1.35 2001/06/25 07:08:56 sjbaker Exp $
+*/
+
 #ifndef _PU_H_
 #define _PU_H_ 1
 
-#include <stdlib.h>
-#include <stdio.h>
-#include <string.h>
-
-#ifdef PU_NOT_USING_GLUT
-#include <windows.h>
-#include <GL/gl.h>
-#include <GL/glu.h>
-#else
-#ifdef WIN32
-#include <windows.h>
-#endif
-#include <GL/glut.h>
-#endif
-
 #include "sg.h"
 #include "fnt.h"
 
-#ifndef TRUE
-#define TRUE  1
-#define FALSE 0
+/*
+  Configuration
+*/
+
+#ifndef PU_NOT_USING_GLUT
+#define _PU_USE_GLUT   1
+#endif
+
+#define _PU_USE_GLUT_FONTS   1
+
+/*
+  Include GLUT
+ */
+
+#ifndef PU_NOT_USING_GLUT
+#  ifdef FREEGLUT_IS_PRESENT
+#    include <GL/freeglut.h>
+#  else
+#    include <GL/glut.h>
+#  endif
 #endif
 
 /*
@@ -57,36 +82,48 @@ typedef void *puFont ;
 #define PUFONT_HELVETICA_12   ((void*)7)
 #define PUFONT_HELVETICA_18   ((void*)8)
 
-#define PU_LEFT_BUTTON		0
-#define PU_MIDDLE_BUTTON	1
-#define PU_RIGHT_BUTTON		2
-#define PU_DOWN			0
-#define PU_UP		        1
+#define PU_LEFT_BUTTON          0
+#define PU_MIDDLE_BUTTON        1
+#define PU_RIGHT_BUTTON         2
+#define PU_DOWN                 0
+#define PU_UP                   1
 
 #else
 
+#ifdef _PU_USE_GLUT_FONTS
+typedef void *GlutFont ;
+#endif
+
 class puFont 
 {
-  void       *glut_font_handle ;
+#ifdef _PU_USE_GLUT_FONTS
+  GlutFont     glut_font_handle ;
+#endif
   fntTexFont * fnt_font_handle ; float pointsize ; float slant ;
 
 public:
 
   puFont ()
   {
+#ifdef _PU_USE_GLUT_FONTS
     glut_font_handle = GLUT_BITMAP_8_BY_13 ;
+#endif
     fnt_font_handle  = NULL ;
   }
 
-  puFont ( void       *gfh )
+#ifdef _PU_USE_GLUT_FONTS
+  puFont ( GlutFont gfh )
   {
     glut_font_handle = gfh  ;
     fnt_font_handle  = NULL ;
   }
+#endif
 
   puFont ( fntTexFont *tfh, float ps, float sl = 0 )
   {
-    glut_font_handle = NULL ;
+#ifdef _PU_USE_GLUT_FONTS
+    glut_font_handle = (GlutFont) 0 ;
+#endif
     fnt_font_handle  = tfh  ;
     pointsize = ps ;
     slant = sl ;
@@ -94,10 +131,10 @@ public:
 
   int getStringDescender () ;
   int getStringHeight    () ;
-  int getStringHeight( char *str ) ;
-  int getStringWidth ( char *str ) ;
+  int getStringHeight( const char *str ) ;
+  int getStringWidth ( const char *str ) ;
 
-  void drawString ( char *str, int x, int y ) ;
+  void drawString ( const char *str, int x, int y ) ;
 } ;
 
 
@@ -124,27 +161,27 @@ extern puFont PUFONT_HELVETICA_18   ;
 #define PU_KEY_GLUT_SPECIAL_OFFSET  256
 
 #ifdef PU_NOT_USING_GLUT
-#define PU_KEY_F1        (1		+ PU_KEY_GLUT_SPECIAL_OFFSET)
-#define PU_KEY_F2        (2		+ PU_KEY_GLUT_SPECIAL_OFFSET)
-#define PU_KEY_F3        (3		+ PU_KEY_GLUT_SPECIAL_OFFSET)
-#define PU_KEY_F4        (4	    + PU_KEY_GLUT_SPECIAL_OFFSET)
-#define PU_KEY_F5        (5		+ PU_KEY_GLUT_SPECIAL_OFFSET)
-#define PU_KEY_F6        (6		+ PU_KEY_GLUT_SPECIAL_OFFSET)
-#define PU_KEY_F7        (7		+ PU_KEY_GLUT_SPECIAL_OFFSET)
-#define PU_KEY_F8        (8		+ PU_KEY_GLUT_SPECIAL_OFFSET)
-#define PU_KEY_F9        (9		+ PU_KEY_GLUT_SPECIAL_OFFSET)
-#define PU_KEY_F10       (10		+ PU_KEY_GLUT_SPECIAL_OFFSET)
-#define PU_KEY_F11       (11		+ PU_KEY_GLUT_SPECIAL_OFFSET)
-#define PU_KEY_F12       (12		+ PU_KEY_GLUT_SPECIAL_OFFSET)
-#define PU_KEY_LEFT      (100		+ PU_KEY_GLUT_SPECIAL_OFFSET)
-#define PU_KEY_UP        (101		+ PU_KEY_GLUT_SPECIAL_OFFSET)
-#define PU_KEY_RIGHT     (102		+ PU_KEY_GLUT_SPECIAL_OFFSET)
-#define PU_KEY_DOWN      (103		+ PU_KEY_GLUT_SPECIAL_OFFSET)
-#define PU_KEY_PAGE_UP   (104		+ PU_KEY_GLUT_SPECIAL_OFFSET)
-#define PU_KEY_PAGE_DOWN (105		+ PU_KEY_GLUT_SPECIAL_OFFSET)
-#define PU_KEY_HOME      (106		+ PU_KEY_GLUT_SPECIAL_OFFSET)
-#define PU_KEY_END       (107		+ PU_KEY_GLUT_SPECIAL_OFFSET)
-#define PU_KEY_INSERT    (108		+ PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_F1        (1             + PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_F2        (2             + PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_F3        (3             + PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_F4        (4             + PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_F5        (5             + PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_F6        (6             + PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_F7        (7             + PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_F8        (8             + PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_F9        (9             + PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_F10       (10            + PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_F11       (11            + PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_F12       (12            + PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_LEFT      (100           + PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_UP        (101           + PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_RIGHT     (102           + PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_DOWN      (103           + PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_PAGE_UP   (104           + PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_PAGE_DOWN (105           + PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_HOME      (106           + PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_END       (107           + PU_KEY_GLUT_SPECIAL_OFFSET)
+#define PU_KEY_INSERT    (108           + PU_KEY_GLUT_SPECIAL_OFFSET)
 
 #else
 #define PU_KEY_F1        (GLUT_KEY_F1        + PU_KEY_GLUT_SPECIAL_OFFSET)
@@ -179,11 +216,16 @@ extern puFont PUFONT_HELVETICA_18   ;
 #define PUARROW_FASTLEFT   6
 #define PUARROW_FASTRIGHT  7
 
-#define PUPLACE_DEFAULT  PUPLACE_RIGHT
-#define PUPLACE_ABOVE    0
-#define PUPLACE_BELOW    1
-#define PUPLACE_LEFT     2
-#define PUPLACE_RIGHT    3
+#define PUPLACE_DEFAULT         PUPLACE_RIGHT
+#define PUPLACE_ABOVE           0
+#define PUPLACE_BELOW           1
+#define PUPLACE_LEFT            2
+#define PUPLACE_RIGHT           3
+#define PUPLACE_CENTERED        4
+#define PUPLACE_TOP_CENTER      5
+#define PUPLACE_BOTTOM_CENTER   6
+#define PUPLACE_LEFT_CENTER     7
+#define PUPLACE_RIGHT_CENTER    8
 
 #define PUCOL_FOREGROUND 0
 #define PUCOL_BACKGROUND 1
@@ -239,10 +281,19 @@ extern int puRefresh ;
 #define PUCLASS_SLIDER           0x00002000
 #define PUCLASS_DIALOGBOX        0x00004000
 #define PUCLASS_ARROW            0x00008000
+#define PUCLASS_LISTBOX          0x00010000
+#define PUCLASS_DIAL             0x00020000
+#define PUCLASS_FILEPICKER       0x00040000
+#define PUCLASS_BISLIDER         0x00080000
+#define PUCLASS_TRISLIDER        0x00100000
+#define PUCLASS_VERTMENU         0x00200000
+#define PUCLASS_LARGEINPUT       0x00400000
 
 /* This function is not required for GLUT programs */
 void puSetWindowSize ( int width, int height ) ;
+void puSetResizeMode ( int mode ) ;
 
+int  puGetWindow       () ;
 int  puGetWindowHeight () ;
 int  puGetWindowWidth  () ;
 
@@ -260,7 +311,19 @@ class puPopupMenu        ;
 class puMenuBar          ;
 class puInput            ;
 class puSlider           ;
+class puListBox          ;
 class puArrowButton      ;
+class puDial             ;
+class puFilePicker       ;
+class puBiSlider         ;
+class puTriSlider        ;
+class puVerticalMenu     ;
+class puLargeInput       ;
+
+// Global function to move active object to the end of the "dlist"
+// so it is displayed in front of everything else
+
+void puMoveToLast ( puObject *ob ) ;
 
 typedef float puColour [ 4 ] ;  /* RGBA */
 
@@ -296,12 +359,20 @@ inline void puSetColour ( puColour c, float r, float g, float b, float a = 1.0f
 
 void  puInit           ( void ) ;
 void  puDisplay        ( void ) ;
+void  puDisplay        ( int window_number ) ;
 int   puMouse          ( int button, int updown, int x, int y ) ;
 int   puMouse          ( int x, int y ) ;
 int   puKeyboard       ( int key, int updown ) ;
 void  puHideCursor     ( void ) ;
 void  puShowCursor     ( void ) ;
 int   puCursorIsHidden ( void ) ;
+void  puDeleteObject   ( puObject *ob ) ;
+
+// Active widget functions
+
+void puDeactivateWidget () ;
+void puSetActiveWidget ( puObject *w, int x, int y ) ;
+puObject *puActiveWidget () ;
 
 /*
   Some deprecated functions for backwards compatibility...
@@ -317,17 +388,17 @@ inline int puGetStringHeight ( puFont fnt )
   return fnt . getStringHeight () ;
 }
 
-inline int puGetStringHeight ( puFont fnt, char *str )
+inline int puGetStringHeight ( puFont fnt, const char *str )
 {
   return fnt . getStringHeight ( str ) ;
 }
 
-inline int puGetStringWidth ( puFont fnt, char *str )
+inline int puGetStringWidth ( puFont fnt, const char *str )
 {
   return fnt . getStringWidth ( str ) ;
 }
 
-inline void puDrawString ( puFont fnt, char *str, int x, int y )
+inline void puDrawString ( puFont fnt, const char *str, int x, int y )
 {
   fnt . drawString ( str, x, y ) ;
 }
@@ -350,8 +421,6 @@ protected:
   void re_eval    () ;
   void update_res () ;
 
-  char *getStringValue () { return res_string ? res_string : string ; }
-
 public:
   puValue ()
   {
@@ -377,9 +446,9 @@ public:
     puRefresh = TRUE ;
   }
 
-  void setValuator ( int   *i ) { res_integer = i ; re_eval () ; }
-  void setValuator ( float *f ) { res_floater = f ; re_eval () ; }
-  void setValuator ( char  *s ) { res_string  = s ; re_eval () ; }
+  void setValuator ( int   *i ) { res_integer = i    ; res_floater = NULL ; res_string = NULL ; re_eval () ; }
+  void setValuator ( float *f ) { res_integer = NULL ; res_floater = f    ; res_string = NULL ; re_eval () ; }
+  void setValuator ( char  *s ) { res_integer = NULL ; res_floater = NULL ; res_string = s    ; re_eval () ; }
 
   void setValue ( int   i ) { integer = i ; floater = (float) i ; sprintf ( string, "%d", i ) ; update_res() ; puRefresh = TRUE ; }
   void setValue ( float f ) { integer = (int) f ; floater = f ; sprintf ( string, "%g", f ) ; update_res() ; puRefresh = TRUE ; }
@@ -407,10 +476,14 @@ public:
   void getValue ( char  *s ) { re_eval () ; strcpy ( s, string ) ; }
 
   int  getValue ( void ) { re_eval () ; return integer ; }
+
+  float getFloatValue () { re_eval () ; return ( floater ) ; }
+  char getCharValue () { re_eval () ; return ( string[0] ) ; }
+  char *getStringValue () { return res_string ? res_string : string ; }
 } ;
 
 typedef void (*puCallback)(class puObject *) ;
-typedef void (*puRenderCallback)(class puObject *, void *) ;
+typedef void (*puRenderCallback)(class puObject *, int dx, int dy, void *) ;
 
 void puSetDefaultStyle ( int  style ) ;
 int  puGetDefaultStyle ( void ) ;
@@ -439,12 +512,15 @@ protected:
   int active      ;
   int highlighted ;
   int am_default  ;
+  int window ;        /* Which window does the object appear in? */
 
-  char *label  ; puFont  labelFont ; int labelPlace ;
-  char *legend ; puFont legendFont ;
+  const char *label  ; puFont  labelFont ; int labelPlace ;
+  const char *legend ; puFont legendFont ; int legendPlace ;
 
   void *user_data ;
   puCallback cb ;
+  puCallback active_cb ;
+  puCallback down_cb ;
   puRenderCallback r_cb ;
   void *render_data ;
 
@@ -453,10 +529,13 @@ protected:
                                                x >= abox.min[0] &&
                                                x <= abox.max[0] &&
                                                y >= abox.min[1] &&
-                                               y <= abox.max[1] ; }
-  virtual void doHit ( int button, int updown, int x, int y ) ;
+                                               y <= abox.max[1] &&
+                                               window == puGetWindow () ; }
+
 
 public:
+  virtual void doHit ( int button, int updown, int x, int y ) ;
+
    puObject ( int minx, int miny, int maxx, int maxy ) ;
   ~puObject () ;
 
@@ -483,7 +562,7 @@ public:
     recalc_bbox() ; puRefresh = TRUE ;
   }
 
-  void setSize ( int w, int h )
+  virtual void setSize ( int w, int h )
   {
     abox.max[0] = abox.min[0] + w ;
     abox.max[1] = abox.min[1] + h ;
@@ -524,6 +603,7 @@ public:
   virtual void draw ( int dx, int dy ) = 0 ;
 
   puGroup     *getParent     ( void ) { return parent ; }
+  void        setParent      ( puGroup* p ) { parent = p ; }
   puObject    *getNextObject ( void ) { return next   ; }
   puObject    *getPrevObject ( void ) { return prev   ; }
 
@@ -531,25 +611,39 @@ public:
   puCallback getCallback ( void )               { return cb ; }
   void       invokeCallback ( void ) { if ( cb ) (*cb)(this) ; }
 
+  void       setActiveCallback ( puCallback c ) { active_cb = c ;    }
+  puCallback getActiveCallback ( void )               { return active_cb ; }
+  void       invokeActiveCallback ( void ) { if ( active_cb ) (*active_cb)(this) ; }
+
+  void       setDownCallback ( puCallback c ) { down_cb = c ;    }
+  puCallback getDownCallback ( void )               { return down_cb ; }
+  virtual void invokeDownCallback ( void ) { if ( down_cb ) (*down_cb)(this) ; }
+
   void       setRenderCallback ( puRenderCallback c, void *d = NULL ) { r_cb = c ; render_data = d ; }
   puRenderCallback getRenderCallback ( void ) { return r_cb ; }
   void      *getRenderCallbackData ( void ) { return render_data ; }
-  void       invokeRenderCallback ( void ) { if ( r_cb ) (*r_cb)(this,render_data) ; }
+  void       invokeRenderCallback ( int dx, int dy ) { if ( r_cb ) (*r_cb)(this, dx, dy, render_data) ; }
 
   void  makeReturnDefault ( int def ) { am_default = def ; }
   int   isReturnDefault   ( void )          { return am_default ; }
 
+  int   getWindow () { return window ; }
+  void  setWindow ( int w ) { window = w ; }
+
   void  setActiveDirn ( int e ) { active_mouse_edge = e ; }
   int   getActiveDirn ( void ) { return active_mouse_edge ; }
 
-  void  setLegend ( char *l ) { legend = l ; recalc_bbox() ; puRefresh = TRUE ; }
-  char *getLegend ( void ) { return legend ; }
+  void  setLegend ( const char *l ) { legend = l ; recalc_bbox() ; puRefresh = TRUE ; }
+  const char *getLegend ( void ) const { return legend ; }
 
   void  setLegendFont ( puFont f ) { legendFont = f ; recalc_bbox() ; puRefresh = TRUE ; }
   puFont getLegendFont ( void ) { return legendFont ; }
 
-  void  setLabel ( char *l ) { label = l ; recalc_bbox() ; puRefresh = TRUE ; }
-  char *getLabel ( void ) { return label ; }
+  void  setLegendPlace ( int lp ) { legendPlace = lp ; recalc_bbox() ; puRefresh = TRUE ; }
+  int   getLegendPlace ( void ) { return legendPlace ; }
+
+  void  setLabel ( const char *l ) { label = l ; recalc_bbox() ; puRefresh = TRUE ; }
+  const char *getLabel ( void ) const { return label ; }
 
   void  setLabelFont ( puFont f ) { labelFont = f ; recalc_bbox() ; puRefresh = TRUE ; }
   puFont getLabelFont ( void ) { return labelFont ; }
@@ -614,7 +708,7 @@ public:
 */
 
 void         puPushLiveInterface        ( puInterface *in ) ;
-void         puPopLiveInterface         ( void ) ;
+void         puPopLiveInterface         ( puInterface *in = 0 ) ;
 int          puNoLiveInterface          ( void ) ;
 puInterface *puGetBaseLiveInterface     ( void ) ;
 puInterface *puGetUltimateLiveInterface ( void ) ;
@@ -634,8 +728,15 @@ protected:
   int num_children ;
   puObject *dlist ;
 
+  int mouse_x ;    // Coordinates of mouse when right button pressed for
+  int mouse_y ;    // drag and drop
+
+  short mouse_active;  // Flag telling whether interface is presently being dragged
+
   void doHit       ( int button, int updown, int x, int y ) ;
 
+  short floating;   // Flag telling whether the interface floats in the window or stays put
+
 public:
 
   puGroup ( int x, int y ) : puObject ( x, y, x, y )
@@ -643,6 +744,10 @@ public:
     type |= PUCLASS_GROUP ;
     dlist = NULL ;
     num_children = 0 ;
+    mouse_x = 0 ;
+    mouse_y = 0 ;
+    mouse_active = FALSE ;
+    floating = FALSE ;
     puPushGroup ( this ) ;
   }
 
@@ -651,6 +756,7 @@ public:
   void recalc_bbox ( void ) ;
   virtual void add    ( puObject *new_object ) ;
   virtual void remove ( puObject *old_object ) ;
+  virtual void empty ( void ) ;
 
   void draw        ( int dx, int dy ) ;
   int  checkHit    ( int button, int updown, int x, int y ) ;
@@ -662,10 +768,13 @@ public:
   virtual void close ( void )
   {
     if ( puGetCurrGroup () != this )
-      fprintf ( stderr, "PUI: puGroup::close() is mismatched!\n" ) ;
+      ulSetError ( UL_WARNING, "PUI: puGroup::close() is mismatched!" ) ;
     else
       puPopGroup () ;
   }
+
+  void setFloating (short value) { floating = value; }
+  short getFloating () { return floating; }
 } ;
 
 
@@ -686,7 +795,12 @@ public:
 class puFrame : public puObject
 {
 protected:
-  virtual int  isHit ( int /* x */, int /* y */ ) { return FALSE ; }
+  int  isHit ( int x, int y ) { return isVisible() && isActive() &&
+                                               x >= abox.min[0] &&
+                                               x <= abox.max[0] &&
+                                               y >= abox.min[1] &&
+                                               y <= abox.max[1] &&
+                                               window == puGetWindow () ; }
 public:
   void draw ( int dx, int dy ) ;
   puFrame ( int minx, int miny, int maxx, int maxy ) :
@@ -694,6 +808,20 @@ public:
   {
     type |= PUCLASS_FRAME ;
   }
+
+  void doHit ( int button, int updown, int x, int y )
+  {
+    if ( puActiveWidget() && ( this != puActiveWidget() ) )
+    {
+      // Active widget exists and is not this one; call its down callback if it exists
+
+      puActiveWidget() -> invokeDownCallback () ;
+      puDeactivateWidget () ;
+    }
+
+    if ( isHit ( x, y ) && ( updown != PU_DRAG ) )
+      puMoveToLast ( this -> parent );
+  }
 } ;
 
 
@@ -717,7 +845,7 @@ protected:
 public:
   void doHit ( int button, int updown, int x, int y ) ;
   void draw  ( int dx, int dy ) ;
-  puButton   ( int minx, int miny, char *l ) :
+  puButton   ( int minx, int miny, const char *l ) :
                  puObject ( minx, miny,
                             minx + puGetStringWidth  ( puGetDefaultLegendFont(), l ) + PUSTR_LGAP + PUSTR_RGAP,
                             miny + puGetStringHeight ( puGetDefaultLegendFont() ) + puGetStringDescender ( puGetDefaultLegendFont() ) + PUSTR_TGAP + PUSTR_BGAP )
@@ -734,7 +862,27 @@ public:
 } ;
 
 
-class puArrowButton : public puButton
+class puOneShot : public puButton
+{
+protected:
+public:
+  void doHit ( int button, int updown, int x, int y ) ;
+
+  puOneShot ( int minx, int miny, const char *l ) : puButton   ( minx, miny, l )
+  {
+    type |= PUCLASS_ONESHOT ;
+  }
+
+  puOneShot ( int minx, int miny, int maxx, int maxy ) :
+                 puButton ( minx, miny, maxx, maxy )
+  {
+    type |= PUCLASS_ONESHOT ;
+  }
+} ;
+
+
+
+class puArrowButton : public puOneShot
 {
 protected:
 
@@ -747,7 +895,7 @@ public:
   void setArrowType ( int i ) { arrow_type = i ; }
 
   puArrowButton ( int minx, int miny, int maxx, int maxy, int ptype ) :
-                 puButton ( minx, miny, maxx, maxy )
+                 puOneShot ( minx, miny, maxx, maxy )
   {
     type |= PUCLASS_ARROW ;
     arrow_type = ptype ;
@@ -755,7 +903,6 @@ public:
 } ;
 
 
-
 class puSlider : public puObject
 {
 protected:
@@ -764,19 +911,38 @@ protected:
   float cb_delta ;
   int   cb_mode ;
   float slider_fraction ;
+  void draw_slider_box ( int dx, int dy, float val, char *box_label = NULL ) ;
 public:
   void doHit ( int button, int updown, int x, int y ) ;
   void draw  ( int dx, int dy ) ;
   puSlider ( int minx, int miny, int sz, int vertical = FALSE ) :
      puObject ( minx, miny, vertical ?
-                             ( minx + puGetStringWidth ( puGetDefaultLegendFont(), "W" ) +
-                                      PUSTR_LGAP + PUSTR_RGAP ) :
+	       ( minx + puGetStringWidth ( puGetDefaultLegendFont(), "W" ) +
+			PUSTR_LGAP + PUSTR_RGAP ) :
+	       ( minx + sz ),
+	      vertical ?
+	       ( miny + sz ) :
+	       ( miny + puGetStringHeight ( puGetDefaultLegendFont() ) +
+			puGetStringDescender ( puGetDefaultLegendFont() ) +
+			PUSTR_TGAP + PUSTR_BGAP )
+	     )
+  {
+    type |= PUCLASS_SLIDER ;
+    slider_fraction = 0.1f ;
+    getValue ( & last_cb_value ) ;
+    vert = vertical ;
+    cb_delta = 0.1f ;
+    cb_mode = PUSLIDER_ALWAYS ;
+  }
+
+  /* Blake Friesen - alternate constructor which lets you explicitly set width */
+  puSlider ( int minx, int miny, int sz, int vertical, int width ) :
+     puObject ( minx, miny, vertical ?
+                             ( minx + width ) :
                              ( minx + sz ),
                             vertical ?
                              ( miny + sz ) :
-                             ( miny + puGetStringHeight ( puGetDefaultLegendFont() ) +
-                                      puGetStringDescender ( puGetDefaultLegendFont() ) +
-                                      PUSTR_TGAP + PUSTR_BGAP )
+                             ( miny + width ) 
                            )
   {
     type |= PUCLASS_SLIDER ;
@@ -800,23 +966,116 @@ public:
 } ;
 
 
-
-class puOneShot : public puButton
+class puBiSlider : public puSlider
 {
+protected:
+  int max_value ;
+  int min_value ;
+
+  int current_max ;
+  int current_min ;
+
+  int active_button ;  // Zero for none, one for min, two for max
+public:
+  void doHit ( int button, int updown, int x, int y ) ;
+  void draw  ( int dx, int dy ) ;
+  puBiSlider ( int minx, int miny, int sz, int vertical = FALSE ) :
+     puSlider ( minx, miny, sz, vertical )
+  {
+    type |= PUCLASS_BISLIDER ;
+    max_value = 1 ;
+    min_value = 0 ;
+    current_max = 1 ;
+    current_min = 0 ;
+    active_button = 0 ;
+  }
+
+  void setMaxValue ( int i )
+  {
+    max_value = i ;
+    slider_fraction = 1.0f / (float)( max_value-min_value+1 ) ;
+  }
+
+  int getMaxValue ( void ) { return max_value ; }
+
+  void setMinValue ( int i )
+  {
+    min_value = i ;
+    slider_fraction = 1.0f / (float)( max_value-min_value+1 ) ;
+  }
+
+  int getMinValue ( void ) { return min_value ; }
+
+  void setCurrentMax ( int i ) { current_max = i ; }
+  int getCurrentMax ( void ) { return current_max ; }
+
+  void setCurrentMin ( int i ) { current_min = i ; }
+  int getCurrentMin ( void ) { return current_min ; }
+
+  void setActiveButton ( int i ) { active_button = i ; }
+  int getActiveButton ( void ) { return active_button ; }
+} ;
+
+
+
+class puTriSlider : public puBiSlider
+{
+protected:
+  // "active_button" is now zero for none, one for min, two for middle, three for max
+  int freeze_ends ;  // true to make end sliders unmovable
+public:
+  void doHit ( int button, int updown, int x, int y ) ;
+  void draw  ( int dx, int dy ) ;
+  puTriSlider ( int minx, int miny, int sz, int vertical = FALSE ) :
+     puBiSlider ( minx, miny, sz, vertical )
+  {
+    type |= PUCLASS_TRISLIDER ;
+    freeze_ends = TRUE ;
+  }
+
+   int getFreezeEnds () {  return freeze_ends ;  }
+   void setFreezeEnds ( int val ) {  freeze_ends = val ;  }
+} ;
+
+
+
+class puListBox : public puButton
+{
+  char ** list ;
+  int num ;
+  int top ;
+
 protected:
 public:
   void doHit ( int button, int updown, int x, int y ) ;
+  void draw  ( int dx, int dy ) ;
+  puListBox  ( int minx, int miny, int maxx, int maxy, char** list = NULL ) ;
 
-  puOneShot ( int minx, int miny, char *l ) : puButton   ( minx, miny, l )
+  void newList     ( char ** _list ) ;
+  int  getNumItems () const { return num ; }
+  int  getTopItem  () const { return top ; }
+  void setTopItem  ( int item_index ) ;
+} ;
+
+
+class puDial : public puSlider
+{
+protected:
+  int wrap ;  // Flag telling whether you can wrap around the bottom of the dial
+public:
+  void doHit ( int button, int updown, int x, int y ) ;
+  void draw  ( int dx, int dy ) ;
+  puDial ( int minx, int miny, int sz ) :
+     puSlider ( minx, miny, sz )
   {
-    type |= PUCLASS_ONESHOT ;
+    type |= PUCLASS_DIAL ;
+    setValue ( 0.0f ) ;
+    wrap = TRUE ;
+    setSize ( sz, sz ) ; /* Override the funky math that the base slider did! */
   }
 
-  puOneShot ( int minx, int miny, int maxx, int maxy ) :
-                 puButton ( minx, miny, maxx, maxy )
-  {
-    type |= PUCLASS_ONESHOT ;
-  }
+  void setWrap ( int in )  {  wrap = in ;  }
+  int getWrap ( void )  {  return wrap ;  }
 } ;
 
 
@@ -832,6 +1091,7 @@ public:
   }
 } ;
 
+
 class puPopupMenu : public puPopup
 {
 protected:
@@ -841,12 +1101,13 @@ public:
     type |= PUCLASS_POPUPMENU ;
   }
 
-  puObject *add_item ( char *str, puCallback cb ) ;
+  puObject *add_item ( const char *str, puCallback cb ) ;
   int  checkHit ( int button, int updown, int x, int y ) ;
   int  checkKey ( int key   , int updown ) ;
   void close ( void ) ;
 } ;
 
+
 class puMenuBar : public puInterface
 {
 protected:
@@ -858,6 +1119,28 @@ public:
     type |= PUCLASS_MENUBAR ;
   }
 
+  void add_submenu ( const char *str, char *items[], puCallback cb[] ) ;
+  void close ( void ) ;
+} ;
+
+
+class puVerticalMenu : public puGroup
+{
+protected:
+public:
+  puVerticalMenu ( int x = -1, int y = -1 ) :
+
+  puGroup ( x < 0 ? puGetWindowWidth() -
+                     ( puGetStringWidth( puGetDefaultLegendFont(), " " )
+                       + PUSTR_TGAP + PUSTR_BGAP ) : x,
+            y < 0 ? puGetWindowHeight() -
+                     ( puGetStringHeight( puGetDefaultLegendFont() )
+                       + PUSTR_TGAP + PUSTR_BGAP ) : y)
+  {
+    type |= PUCLASS_VERTMENU ;
+    floating = TRUE ;
+  }
+
   void add_submenu ( char *str, char *items[], puCallback cb[] ) ;
   void close ( void ) ;
 } ;
@@ -869,8 +1152,10 @@ class puInput : public puObject
   int cursor_position ;
   int select_start_position ;
   int select_end_position ;
+  char *valid_data ;
 
   void normalize_cursors ( void ) ;
+  void removeSelectRegion ( void ) ;
 
 public:
   void draw     ( int dx, int dy ) ;
@@ -902,6 +1187,40 @@ public:
     if ( e ) *e = select_end_position   ;
   }
 
+  char *getValidData () { return valid_data ; }
+  void setValidData ( char *data )
+  {
+    if ( valid_data )
+    {
+      delete valid_data ;
+      valid_data = NULL ;
+    }
+
+    if ( data )
+    {
+      valid_data = new char [ strlen ( data ) + 1 ] ;
+      strcpy ( valid_data, data ) ;
+    }
+  }
+
+  void addValidData ( char *data )
+  {
+    int new_data_len = 1 ;
+    if ( valid_data ) new_data_len += strlen ( valid_data ) ;
+    if ( data )       new_data_len += strlen ( data ) ;
+    char *new_data = new char [ new_data_len ] ;
+    strcpy ( new_data, "\0" ) ;
+    if ( valid_data ) strcat ( new_data, valid_data ) ;
+    if ( data )       strcat ( new_data, data ) ;
+    delete valid_data ;
+    valid_data = new_data ;
+  }
+
+  int isValidCharacter ( char c )
+  {
+    return ( ( strchr ( valid_data, c ) != NULL ) ? 1 : 0 ) ;
+  }
+
   puInput ( int minx, int miny, int maxx, int maxy ) :
              puObject ( minx, miny, maxx, maxy )
   {
@@ -913,8 +1232,22 @@ public:
     select_start_position = -1 ;
     select_end_position   = -1 ;
 
+    valid_data = NULL ;
+
     setColourScheme ( 0.8f, 0.7f, 0.7f ) ; /* Yeukky Pink */
   }
+
+  ~puInput ()
+  {
+    if ( valid_data ) free ( valid_data ) ;
+  }
+
+  virtual void invokeDownCallback ( void )
+  {
+    rejectInput () ;
+    normalize_cursors () ;
+    if ( down_cb ) (*down_cb)(this) ;
+  }
 } ;
 
 
@@ -938,7 +1271,6 @@ public:
 } ;
 
 
-
 class puDialogBox : public puPopup
 {
 protected:
@@ -950,5 +1282,151 @@ public:
   }
 } ;
 
-#endif
 
+class puFilePicker : public puDialogBox
+{
+  char** files ;
+  char*  dflag ;
+  int num_files   ;
+  int arrow_count ;
+
+  char startDir [ PUSTRING_MAX ] ;
+
+  void find_files () ;
+  static void handle_select ( puObject* ) ;
+  static void input_entered ( puObject* ) ;
+
+  puFrame   *frame         ;
+  puListBox *list_box      ;
+  puSlider  *slider        ;
+  puOneShot *cancel_button ;
+  puOneShot *ok_button     ;
+/* puInput   *input         ; */
+
+protected:
+  void puFilePickerInit ( int x, int y, int w, int h,
+                          int arrows, const char *dir, const char *title ) ;
+
+public:
+
+  puFilePicker ( int x, int y, int w, int h, int arrows,
+                 const char *dir, const char *title = "Pick a file" ) ;
+  puFilePicker ( int x, int y, int w, int h,
+                 const char *dir, const char *title = "Pick a file" ) ;
+  puFilePicker ( int x, int y, int arrows,
+                 const char* dir, const char *title = "Pick a file" ) ;
+  puFilePicker ( int x, int y,
+                 const char* dir, const char *title = "Pick a file" ) ;
+
+  ~puFilePicker () ;
+
+  /* Not for application use!! */
+/* puInput *__getInput () { return input ; } */
+
+  void setSize ( int w, int h ) ;
+} ;
+
+
+
+class puLargeInput : public puGroup
+{
+  int num_lines ;              // Number of lines of text in the box
+  int lines_in_window ;        // Number of lines showing in the window
+  int top_line_in_window ;     // Number of the first line in the window
+  int max_width ;              // Width of longest line of text in box, in pixels
+  int slider_width ;
+  int accepting ;
+  int cursor_position ;
+  int select_start_position ;
+  int select_end_position ;
+  char *valid_data ;
+
+  puSlider *bottom_slider ;    // Horizontal slider at bottom of window
+  puSlider *right_slider ;     // Vertical slider at right of window
+
+  char *text ;                 // Pointer to text in large input box
+
+  short arrow_count ;          // Number of up/down arrows above and below the right slider
+
+  void normalize_cursors ( void ) ;
+  void removeSelectRegion ( void ) ;
+
+public:
+  puLargeInput ( int x, int y, int w, int h, int arrows, int sl_width ) ;
+  ~puLargeInput () ;
+
+  void setSize ( int w, int h ) ;
+
+  int getNumLines () {  return num_lines ;  }
+  void setTopLineInWindow ( int val ) {  top_line_in_window = val ;  }
+
+  void draw     ( int dx, int dy ) ;
+  int  checkHit ( int button, int updown, int x, int y ) ;
+  void doHit    ( int button, int updown, int x, int y ) ;
+  int  checkKey ( int key, int updown ) ;
+
+  void setSelectRegion ( int s, int e ) ;
+  void selectEntireLine ( void ) ;
+
+  int  isAcceptingInput ( void ) { return accepting ; }
+  void rejectInput      ( void ) { accepting = FALSE ; }
+
+  void acceptInput ( void )
+  {
+    accepting = TRUE ;
+    cursor_position = strlen ( getStringValue() ) ;
+    select_start_position = select_end_position = -1 ;
+  }
+
+  int  getCursor ( void )  { return cursor_position ; }
+  void setCursor ( int c ) { cursor_position = c ; }
+
+  void getSelectRegion ( int *s, int *e )
+  {
+    if ( s ) *s = select_start_position ;
+    if ( e ) *e = select_end_position   ;
+  }
+
+  char *getValidData () { return valid_data ; }
+  void setValidData ( char *data )
+  {
+    if ( valid_data )
+    {
+      free ( valid_data ) ;
+      valid_data = NULL ;
+    }
+
+    if ( data )
+    {
+      valid_data = new char[strlen(data)+1] ;
+      strcpy ( valid_data, data ) ;
+    }
+  }
+
+  void addValidData ( char *data )
+  {
+    int new_data_len = 1 ;
+    if ( valid_data ) new_data_len += strlen ( valid_data ) ;
+    if ( data )       new_data_len += strlen ( data ) ;
+    char*new_data = new char[new_data_len] ;
+    strcpy ( new_data, "\0" ) ;
+    if ( valid_data ) strcat ( new_data, valid_data ) ;
+    if ( data )       strcat ( new_data, data ) ;
+    delete valid_data ;
+    valid_data = new_data ;
+  }
+
+  int isValidCharacter ( char c )
+  {
+    return ( ( strchr ( valid_data, c ) != NULL ) ? 1 : 0 ) ;
+  }
+
+  void  setText ( char *l ) ;
+  char *getText ( void ) { return text ; }
+  void  addNewLine ( char *l ) ;
+  void  addText ( char *l ) ;
+  void  appendText ( char *l ) ;
+  void  removeText ( int start, int end ) ;
+} ;
+
+#endif
